use super::ApiError;
use anyhow::Context;
use axum::{
    async_trait,
    body::BoxBody,
    extract::{FromRequest, FromRequestParts},
    http::Request,
    response::IntoResponse,
};
use jwt_compact::{
    alg::{SigningKey, VerifyingKey},
    AlgorithmSignature, ParseError, Token, UntrustedToken, ValidationError,
};
use jwt_compact::{Algorithm, AlgorithmExt};
use serde::{de::DeserializeOwned, Serialize};
use ssh_key::{Certificate, SshSig};
use std::marker::PhantomData;
use std::{fmt::Debug, ops::Deref};
use tracing::trace;

#[derive(Debug, Clone)]
pub struct CertificateBody(pub Certificate);

// we must implement `FromRequest` (and not `FromRequestParts`) to consume the body
#[async_trait]
impl<S> FromRequest<S, BoxBody> for CertificateBody
where
    S: Send + Sync,
{
    type Rejection = ApiError;

    async fn from_request(req: Request<BoxBody>, state: &S) -> Result<Self, Self::Rejection> {
        let body = String::from_request(req, state)
            .await
            .context("failed to extract body")?;

        let cert = Certificate::from_openssh(&body)
            .with_context(|| format!("failed to parse '{}'", body))?;
        trace!(%body, "extracted certificate");
        Ok(Self(cert))
    }
}

#[derive(Debug, Clone)]
pub struct SignatureBody(pub SshSig);

#[async_trait]
impl<S> FromRequest<S, BoxBody> for SignatureBody
where
    S: Send + Sync,
{
    type Rejection = ApiError;

    async fn from_request(req: Request<BoxBody>, state: &S) -> Result<Self, Self::Rejection> {
        let body = String::from_request(req, state)
            .await
            .context("failed to extract body")?;

        let sig = SshSig::from_pem(&body).with_context(|| format!("failed to parse '{}'", body))?;
        trace!(%body, "extracted signature");
        Ok(Self(sig))
    }
}

pub trait AsJWTVerifier: Send + Sync {
    type Algo: Algorithm + Default;
    fn as_secret(&self) -> &<Self::Algo as Algorithm>::VerifyingKey;
}

pub struct JWTString(String);

impl From<String> for JWTString {
    fn from(s: String) -> Self {
        Self(s)
    }
}

#[derive(Debug)]
pub struct JWTAuthenticated<T, S, Q, E> {
    pub data: T,
    _marker: PhantomData<(Q, S, E)>,
}

impl<T, S, Q, E> From<T> for JWTAuthenticated<T, S, Q, E> {
    fn from(data: T) -> Self {
        Self {
            data,
            _marker: Default::default(),
        }
    }
}

impl<T, S, Q, E> Deref for JWTAuthenticated<T, S, Q, E> {
    type Target = T;
    fn deref(&self) -> &Self::Target {
        &self.data
    }
}

impl<
        T: Serialize + DeserializeOwned + Clone + Debug,
        S: AsJWTVerifier,
        Q: FromRequestParts<S> + Debug + Into<JWTString>,
        E: From<<Q as FromRequestParts<S>>::Rejection>
            + From<ValidationError>
            + From<ParseError>
            + Debug
            + Send
            + Sync,
    > JWTAuthenticated<T, S, Q, E>
{
    pub fn new(data: T) -> Self {
        Self {
            data,
            _marker: Default::default(),
        }
    }
}

#[async_trait]
impl<
        T: Serialize + DeserializeOwned + Clone + Debug,
        S: AsJWTVerifier,
        Q: FromRequestParts<S> + Debug + Into<JWTString>,
        E: From<<Q as FromRequestParts<S>>::Rejection>
            + From<ValidationError>
            + From<ParseError>
            + Debug
            + Send
            + Sync
            + IntoResponse,
    > FromRequestParts<S> for JWTAuthenticated<T, S, Q, E>
{
    type Rejection = E;

    async fn from_request_parts(
        parts: &mut axum::http::request::Parts,
        state: &S,
    ) -> Result<Self, Self::Rejection> {
        let JWTString(token) = Q::from_request_parts(parts, state).await?.into();
        let token = UntrustedToken::new(&token)?;
        let verified: Token<T> =
            <S::Algo as Default>::default().validate_integrity(&token, &state.as_secret())?;
        Ok(Self {
            data: verified.claims().custom.clone(),
            _marker: Default::default(),
        })
    }
}
